-- TODO - Easier accessing of information, needs sane getters

local excitingBrickGamePlayfield = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_playfield")
local excitingBrickGameBag = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_bag")
local excitingBrickGameMovement = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_piece_movement")
local excitingBrickGameSpeedCurve = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_speed_curve")
local excitingBrickGameGarbageManager = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_playfield_garbage_manager")
local excitingBrickGameScoreCalculator = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_score_calculator")
local excitingBrickShapes = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_default_exciting_brick_shapes")

local utils = require("mods").requireFromPlugin("utils")
local virtualGamepad = require("mods").requireFromPlugin("virtual_gamepad")

local excitingBrickShapeStruct = require("mods").requireFromPlugin("structs.exciting_brick_shape")

local excitingBrickGameGame = {}

excitingBrickGameGame._MT = {}

function excitingBrickGameGame._MT:__index(key)
    if excitingBrickGameGame._defaultValues[key] then
        return excitingBrickGameGame._defaultValues[key]
    end

    return rawget(self, key)
end

excitingBrickGameGame._defaultValues = {}

--[[ Events ]]
-- All events starts with event name then unpaused time of event

-- Events TODO:
--- Lines cleared
--- Garbage received
--- Level changed
--- Topped up

function excitingBrickGameGame._defaultValues:queueRotationEvent(direction)
    local excitingBrickShape = self._currentexcitingBrickShape

    self:queueEvent("excitingBrickShapeRotate", self._timeUnpaused, excitingBrickShape.x, excitingBrickShape.y, excitingBrickShape.rotation, direction)
end

function excitingBrickGameGame._defaultValues:queueMovementEvent(offsetX, offsetY, softDrop, hardDrop)
    local excitingBrickShape = self._currentexcitingBrickShape

    self:queueEvent("excitingBrickShapeMove", self._timeUnpaused, excitingBrickShape.x, excitingBrickShape.y, offsetX, offsetY, softDrop or self._softDropping, hardDrop or self._hardDropping)
end

function excitingBrickGameGame._defaultValues:queueNextexcitingBrickShapeEvent(next, given)
    local givenName = given and given.name

    self:queueEvent("excitingBrickShapeNext", self._timeUnpaused, next, givenName)
end

function excitingBrickGameGame._defaultValues:queueHoldexcitingBrickShapeEvent(held)
    self:queueEvent("excitingBrickShapeHold", self._timeUnpaused, held)
end

function excitingBrickGameGame._defaultValues:queueexcitingBrickShapeLockDownEvent()
    self:queueEvent("excitingBrickShapeLockDown", self._timeUnpaused)
end

-- TSpin values - false, "mini", "proper"
function excitingBrickGameGame._defaultValues:queueTSpinEvent(value)
    self:queueEvent("updateTSpinLevel", self._timeUnpaused, value or self._tSpinValid)
end

function excitingBrickGameGame._defaultValues:toppedOut()
    self._pauseEnabled = false
    self._toppedOut = true
    self._paused = true
end

function excitingBrickGameGame._defaultValues:canUnpause()
    return self._pauseEnabled
end

-- TODO - Better fix for gamepad pause being stuck
function excitingBrickGameGame._defaultValues:pause()
    if self._pauseEnabled then
        self._paused = true
        self._gamepad._buttons.pause = false
    end
end

-- TODO - Better fix for gamepad pause being stuck
function excitingBrickGameGame._defaultValues:unpause()
    if self._pauseEnabled then
        self._paused = false
        self._gamepad._buttons.pause = false
    end
end

function excitingBrickGameGame._defaultValues:togglePause()
    if self._paused then
        self:unpause()

    else
        self:pause()
    end
end

-- TODO - Simplify?
function excitingBrickGameGame._defaultValues:updateLockDownStatus(canMoveDown)
    if canMoveDown then
        if self._lockDownActive then
            self._lockDownActive = false
            self._lockDownTimer = 0
        end

    else
        if not self._lockDownActive then
            self._lockDownActive = true
            self._lockDownTimer = 0
        end
    end
end

function excitingBrickGameGame._defaultValues:getRelevantGravityLineTime()
    return self._softDropping and self._gravityLineTimeSoftDrop or self._gravityLineTime
end

function excitingBrickGameGame._defaultValues:setRelevantGravityLineTime(time)
    if self._softDropping then
        self._gravityLineTimeSoftDrop = time

    else
        self._gravityLineTime = time
    end
end

function excitingBrickGameGame._defaultValues:awardScoreDroppingBonus(rows, softDrop, hardDrop)
    local dropBonus = self._scoreCalculator:getDropBonus(softDrop or self._softDropping, hardDrop or self._hardDropping)

    self._score = self._score + dropBonus
end

-- TODO - Handle lockdown sanely
function excitingBrickGameGame._defaultValues:updateGravity(dt)
    local softDrop, hardDrop = self._softDropping, self._hardDropping

    local secondsPerRow = hardDrop and self._hardDropSpeedDuration or self._speedSecondsPerRow

    if softDrop and not hardDrop then
        secondsPerRow = secondsPerRow * self._softDropSpeedMultiplier
    end

    local lineTime = self:getRelevantGravityLineTime()

    lineTime = lineTime + dt
    self._gravityTotalTime = self._gravityTotalTime + dt

    if self._currentexcitingBrickShape then
        while lineTime >= secondsPerRow do
            lineTime = lineTime - secondsPerRow

            local moved = self:moveexcitingBrickShapeDown()
            local canMoveAgain = self:canMoveexcitingBrickShapeDown()

            self:updateLockDownStatus(canMoveAgain)
        end
    end

    self:setRelevantGravityLineTime(lineTime)
end

function excitingBrickGameGame._defaultValues:updateTimers(dt)
    self._timeTotal = self._timeTotal + dt

    if not self._paused then
        self._timeUnpaused = self._timeUnpaused + dt
    end
end

function excitingBrickGameGame._defaultValues:clearLinesName(rows, allClear)
    return excitingBrickGameMovement.lineClearName(#rows, self._tSpinValid, allClear)
end

-- Clearing no lines does not modify BtB bonus
-- Any amount of cleared lines with a T-Spin keeps the bonus
-- Only a Exciting! clear keeps the bonus otherwise
function excitingBrickGameGame._defaultValues:updateBackToBack(rows)
    if #rows == 0 and not self._tSpinValid then
        return
    end

    local currentBonus = self._backToBackBonus

    -- Both Mini and proper T-Spins count
    self._backToBackBonus = self._tSpinValid or #rows == 4

    return currentBonus and self._backToBackBonus
end

-- The initial clear does not count as combo, only subsequent clears
-- Clearing 3 times in a row is a combo of 2
function excitingBrickGameGame._defaultValues:updateCombo(rows)
    if #rows > 0 then
        self._comboCount = self._comboCount + 1
        self._comboActive = true

        return self._comboCount - 1

    else
        self:resetCombo()

        return self._combo or 0
    end
end

function excitingBrickGameGame._defaultValues:updateTSpin(offsetX, offsetY, offsetIndex, direction)
    local spinLevel = excitingBrickGameMovement.checkTSpinLevel(self._currentexcitingBrickShape, self._playfield, offsetX, offsetY, offsetIndex, direction)

    if spinLevel then
        self:queueTSpinEvent()
    end

    self._tSpinValid = spinLevel
end

function excitingBrickGameGame._defaultValues:getGameDuration()
    return self._timeUnpaused
end

function excitingBrickGameGame._defaultValues:getLineClearInformations()
    return self._lineClearInformations
end

function excitingBrickGameGame._defaultValues:addLineClearInformation(name, linesCleared, tSpin, awardBackToBack, combo, allClear, time)
    -- Ignore if clear doesn't have a proper name
    if not name then
        return
    end

    local info = {
        name = name,
        rows = linesCleared,
        lines = #linesCleared,
        tSpin = tSpin,
        backToBack = awardBackToBack,
        combo = combo,
        allClear = allClear,
        time = time or self._timeUnpaused
    }

    table.insert(self._lineClearInformations, info)
end

function excitingBrickGameGame._defaultValues:clearLines()
    local clearedLines, allClear = excitingBrickGameMovement.findFilledLines(self._playfield)
    local clearedLinesCount = #clearedLines
    local clearName = self:clearLinesName(clearedLines, allClear)

    excitingBrickGameMovement.clearLines(self._playfield, clearedLines)

    local currentCombo = self:updateCombo(clearedLines)
    local awardBackToBack = self:updateBackToBack(clearedLines)

    self._linesCleared = self._linesCleared + clearedLinesCount

    if clearName then
        local garbageLines = self._garbageManager:queueOutboundGarbage(clearedLinesCount, self._tSpinValid, awardBackToBack, currentCombo, allClear)
        local score = self._scoreCalculator:calculateScore(self._level, clearedLinesCount, self._tSpinValid, awardBackToBack, currentCombo, allClear)

        self:addLineClearInformation(clearName, clearedLines, self._tSpinValid, awardBackToBack, currentCombo)

        self._linesGenerated = self._linesGenerated + garbageLines
        self._score = self._score + score
    end
end

function excitingBrickGameGame._defaultValues:addAllInboundGarbageToPlayfield()
    self._garbageManager:addAllInboundGarbageToPlayfield(self._playfield)
end

-- TODO - Handle lockdown sanely
function excitingBrickGameGame._defaultValues:updateLockDown(dt)
    self:updateLockDownStatus(self:canMoveexcitingBrickShapeDown())

    if self._lockDownActive then
        self._lockDownTimer = self._lockDownTimer + dt

        if self._hardDropping or self._lockDownTimer >= self._lockDownDuration then
            self:lockDownexcitingBrickShape()
            self:clearLines()
            self:addAllInboundGarbageToPlayfield()
            self:resetHardDrop()
            self:nextexcitingBrickShape()
            self:resetHolding()
        end
    end
end

function excitingBrickGameGame._defaultValues:getCurrentDASMovementOffset()
    local button = self._dasDirectionButton

    if button == "left" then
        return -1

    elseif button == "right"then
        return 1

    else
        return 0
    end
end

function excitingBrickGameGame._defaultValues:setDASButton(button)
    self._dasDirectionButton = button
    self._dasActive = false
    self._dasDuration = 0
end

function excitingBrickGameGame._defaultValues:updateDAS(dt)
    local dasButtonHeld = self._gamepad:getButton(self._dasDirectionButton)

    if dasButtonHeld then
        self._dasDuration = self._dasDuration + dt

        if not self._dasActive then
            if self._dasDuration >= self._dasActivationDelayDuration then
                self._dasDuration = self._dasDuration - self._dasActivationDelayDuration
                self._dasActive = true
            end
        end

        if self._dasActive then
            local dasMovementOffsetX = self:getCurrentDASMovementOffset()

            while self._dasDuration >= self._dasActiveSpeedDuration do
                self._dasDuration = self._dasDuration - self._dasActiveSpeedDuration

                self:moveexcitingBrickShape(dasMovementOffsetX, 0)
            end
        end
    end
end

function excitingBrickGameGame._defaultValues:update(dt)
    if not self._paused then
        self:updateGravity(dt)
        self:updateLockDown(dt)
        self:updateDAS(dt)
    end

    self:updateTimers(dt)
end

function excitingBrickGameGame._defaultValues:resetCombo()
    self._comboCount = 0
    self._comboActive = false
end

function excitingBrickGameGame._defaultValues:resetTSpin()
    self._tSpinValid = false
end

function excitingBrickGameGame._defaultValues:resetLockDownIfPossible()
    if self._lockDownActive then
        self._lockDownMoves = self._lockDownMoves + 1

        if self._lockDownMoves <= self._lockDownMoveLimit then
            self._lockDownTimer = 0
            self:updateLockDownStatus(self:canMoveexcitingBrickShapeDown())

        else
            self._lockDownRotationsDisabled = true
        end
    end
end

-- T Spin is reset on any movement, regardless of it being player input or gravity
function excitingBrickGameGame._defaultValues:moveexcitingBrickShape(offsetX, offsetY)
    local moved = excitingBrickGameMovement.moveexcitingBrickShape(self._currentexcitingBrickShape, self._playfield, offsetX, offsetY)

    if moved then
        self:updateGhostexcitingBrickShape()
        self:resetTSpin()
        self:queueMovementEvent(offsetX, offsetY)

        if offsetY > 0 then
            self:awardScoreDroppingBonus(offsetY)
        end

        -- Only applies to horizontal movement
        if offsetX ~= 0 then
            self:resetLockDownIfPossible()
        end
    end

    return moved
end

-- Short hand for moveexcitingBrickShape with (0, 1)
function excitingBrickGameGame._defaultValues:moveexcitingBrickShapeDown()
    return self:moveexcitingBrickShape(0, 1)
end

function excitingBrickGameGame._defaultValues:canMoveexcitingBrickShape(offsetX, offsetY)
    return excitingBrickGameMovement.canMoveexcitingBrickShape(self._currentexcitingBrickShape, self._playfield, offsetX, offsetY)
end

-- Short hand for canMoveexcitingBrickShape with (0, 1)
function excitingBrickGameGame._defaultValues:canMoveexcitingBrickShapeDown()
    return self:canMoveexcitingBrickShape(0, 1)
end

-- T Spin status is set if at least three out of four diagonal corners are solid
function excitingBrickGameGame._defaultValues:rotateexcitingBrickShape(direction)
    if self._lockDownRotationsDisabled and not self:canMoveexcitingBrickShapeDown() then
        return
    end

    local rotated, offsetX, offsetY, offsetIndex = excitingBrickGameMovement.rotateexcitingBrickShape(self._currentexcitingBrickShape, self._playfield, direction)

    if rotated then
        self:updateGhostexcitingBrickShape()
        self:resetLockDownIfPossible()
        self:updateTSpin(offsetX, offsetY, offsetIndex, direction)
        self:queueRotationEvent(direction)
    end

    return rotated
end

function excitingBrickGameGame._defaultValues:updateGhostexcitingBrickShape()
    self._ghostexcitingBrickShape.x = self._currentexcitingBrickShape.x
    self._ghostexcitingBrickShape.y = self._currentexcitingBrickShape.y
    self._ghostexcitingBrickShape.rotation = self._currentexcitingBrickShape.rotation

    while excitingBrickGameMovement.moveexcitingBrickShape(self._ghostexcitingBrickShape, self._playfield, 0, 1) do

    end
end

function excitingBrickGameGame._defaultValues:lockDownexcitingBrickShape()
    excitingBrickGameMovement.lockDownexcitingBrickShape(self._currentexcitingBrickShape, self._playfield)

    self:queueexcitingBrickShapeLockDownEvent()
end

function excitingBrickGameGame._defaultValues:holdexcitingBrickShape()
    if self._holdingCounter < self._holdingPerPlacement then
        local heldexcitingBrickShape = self._heldexcitingBrickShape

        self._heldexcitingBrickShape = self._currentexcitingBrickShape.name
        self._holdingCounter = self._holdingCounter + 1

        self:queueHoldexcitingBrickShapeEvent(self._currentexcitingBrickShape.name)
        self:nextexcitingBrickShape(heldexcitingBrickShape)
    end
end

function excitingBrickGameGame._defaultValues:resetLockDown()
    self._lockDownActive = false
    self._lockDownRotationsDisabled = false
    self._lockDownTimer = 0
    self._lockDownMoves = 0
end

function excitingBrickGameGame._defaultValues:resetHardDrop()
    self._hardDropping = false
end

function excitingBrickGameGame._defaultValues:resetHolding()
    self._holdingCounter = 0
end

function excitingBrickGameGame._defaultValues:resetDAS()
    self._dasActive = false
    self._dasDirectionButton = "none"
    self._dasDuration = 0
end

-- Spawn the top preview excitingBrickShape, or use provided excitingBrickShape (from hold)
-- Tops off if the initial spawn or row below collides with solids
-- TODO - Trigger topped out if piece gets stuck above visible board
function excitingBrickGameGame._defaultValues:nextexcitingBrickShape(excitingBrickShape)
    excitingBrickShape = excitingBrickShape or self._pieceBag:pop()

    if type(excitingBrickShape) == "string" then
        self._currentexcitingBrickShape = excitingBrickShapeStruct.createexcitingBrickShape(excitingBrickShapes[excitingBrickShape])
        self._ghostexcitingBrickShape = excitingBrickShapeStruct.createexcitingBrickShape(excitingBrickShapes[excitingBrickShape])

    else
        self._currentexcitingBrickShape = excitingBrickShapeStruct.createexcitingBrickShape(excitingBrickShape)
        self._ghostexcitingBrickShape = excitingBrickShapeStruct.createexcitingBrickShape(excitingBrickShape)
    end

    self._currentexcitingBrickShape.x = self._exciting_brick_shapeInitialX + self._currentexcitingBrickShape.spawnOffsetX
    self._currentexcitingBrickShape.y = self._exciting_brick_shapeInitialY + self._currentexcitingBrickShape.spawnOffsetY

    local spawnValid = not excitingBrickGameMovement.exciting_brick_shapeColliding(self._currentexcitingBrickShape, self._playfield)
    local rowBelowValid = self:moveexcitingBrickShapeDown()

    if not spawnValid or not rowBelowValid then
        self:toppedOut()

        return false
    end

    self:resetLockDown()
    self:queueNextexcitingBrickShapeEvent(self._currentexcitingBrickShape.name, excitingBrickShape)

    return true
end

function excitingBrickGameGame._defaultValues:updateSpeed(level)
    level = level or self._level

    self._speedSecondsPerRow = self._speedCurve:timePerRow(level)
end

function excitingBrickGameGame._defaultValues:setLevel(level)
    self._level = level

    self:updateSpeed(level)
end

function excitingBrickGameGame._defaultValues:getLevel(level)
    return self._level
end

function excitingBrickGameGame._defaultValues:queueEvent(...)
    table.insert(self._events, {...})
end

function excitingBrickGameGame._defaultValues:pullEvent()
    return table.remove(self._events, 1)
end

function excitingBrickGameGame._defaultValues:hasEvents()
    return #self._events > 0
end

function excitingBrickGameGame._defaultValues:gamepadInit()
    self._gamepad._buttonPressed = function(gamepad, button)
        return self:gamepadButtonPressed(gamepad, button)
    end
    self._gamepad._buttonReleased = function(gamepad, button)
        return self:gamepadButtonReleased(gamepad, button)
    end
end

function excitingBrickGameGame._defaultValues:gamepadButtonPressed(gamepad, button)
    if button == "pause" then
        self:togglePause()
    end

    if self._paused then
        return
    end

    if button == "left" then
        self:moveexcitingBrickShape(-1, 0)
        self:setDASButton(button)

    elseif button == "right" then
        self:moveexcitingBrickShape(1, 0)
        self:setDASButton(button)

    elseif button == "softDrop" then
        self._softDropping = true

    elseif button == "hardDrop" then
        self._hardDropping = true

    elseif button == "hold" then
        self:holdexcitingBrickShape()

    elseif button == "rotateLeft" then
        self:rotateexcitingBrickShape(-1)

    elseif button == "rotateRight" then
        self:rotateexcitingBrickShape(1)
    end
end

function excitingBrickGameGame._defaultValues:gamepadButtonReleased(gamepad, button)
    if button == self._dasDirectionButton then
        self:setDASButton(nil)

    elseif button == "softDrop" then
        self._softDropping = false
    end
end

-- TODO - Move default settings to own file
-- Current defaults are trying to mimic PPT
function excitingBrickGameGame.createGame(settings)
    local data = {}

    settings = settings or {}

    -- Settings
    data._exciting_brick_shapeInitialX = settings.exciting_brick_shapeInitialX or 3
    data._exciting_brick_shapeInitialY = settings.exciting_brick_shapeInitialY or 18

    data._playfield = excitingBrickGamePlayfield.createPlayfield(settings.width, settings.height, settings.visibleWidth, settings.visibleHeight)
    data._pieceBag = excitingBrickGameBag.createBag(settings.mode or "seven_bag")
    data._speedCurve = excitingBrickGameSpeedCurve.createSpeedCurve(settings.speedFormula, settings.speedLookup)

    data._lockDownDuration = settings.lockDownDuration or 0.5
    data._lockDownMode = settings.lockDownMode or "moveRotationReset"
    data._lockDownMoveLimit = settings.lockDownMoveLimit or 15

    data._softDropSpeedMultiplier = settings.softDropSpeedMultiplier or (1 / 20)
    data._hardDropSpeedDuration = settings.hardDropSpeedDuration or (1 / 1200) -- 20 lines per frame at 60fps
    data._dasActiveSpeedDuration = settings.dasActiveSpeedDuration or 0.05

    data._exciting_brick_shapeSpawnDelayDuration = settings.exciting_brick_shapeSpawnDelayDuration or (7 / 60)
    data._dasActivationDelayDuration = settings.dasActivationDelayDuration or (9 / 60)

    data._holdingPerPlacement = settings.holdingPerPlacement or 1

    data._gamepad = virtualGamepad.createGamepad()

    data._garbageManager = excitingBrickGameGarbageManager.createGarbageManager()
    data._scoreCalculator = excitingBrickGameScoreCalculator.createScoreCalculator()

    data._level = settings.level or 1

    data._events = {}

    -- Gameplay logic
    data._gravityTotalTime = 0
    data._gravityLineTime = 0
    data._gravityLineTimeSoftDrop = 0

    data._toppedOut = false
    data._paused = false
    data._pauseEnabled = true

    data._timeTotal = 0
    data._timeUnpaused = 0

    data._linesCleared = 0
    data._linesGenerated = 0
    data._score = 0
    data._lineClearInformations = {}

    data._comboCount = 0
    data._comboActive = false
    data._backToBackBonus = false

    data._softDropping = false
    data._hardDropping = false

    data._dasActive = false
    data._dasDirectionButton = "none"

    --Exciting Brick Shape fields
    data._currentexcitingBrickShape = nil
    data._ghostexcitingBrickShape = nil
    data._heldexcitingBrickShape = nil

    local self = setmetatable(data, excitingBrickGameGame._MT)

    self:gamepadInit()
    self:nextexcitingBrickShape()

    self:resetCombo()
    self:resetTSpin()
    self:resetLockDown()
    self:resetHolding()
    self:resetDAS()

    self:setLevel(data._level)

    return self
end

return excitingBrickGameGame