local excitingBrickGameTheme = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_theme")
local excitingBrickGameMode = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_mode")
local excitingBrickGameGame = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_game")

local inputHandler = require("mods").requireFromPlugin("input_handler")

local gameHandler = {}

gameHandler._MT = {}

function gameHandler._MT:__index(key)
    if gameHandler._defaultValues[key] then
        return gameHandler._defaultValues[key]
    end

    return rawget(self, key)
end

gameHandler._defaultValues = {}

function gameHandler._defaultValues:draw()
    love.graphics.push()

    for _, game in ipairs(self._games) do
        self._theme:drawGame(self._gameMode, game)

        love.graphics.translate(600, 0)
    end

    love.graphics.pop()
end

-- TODO - Handle this better when modes are better implemented
function gameHandler._defaultValues:update(dt)
    self._gameMode:update(dt)
    self._gameMode:updateGames(dt, self._games)
    self._paused = false

    for _, game in ipairs(self._games) do
        self._gameMode:updateGame(dt, game)
        self._paused = self._paused or game._paused
    end
end

function gameHandler._defaultValues:updateVirtualGamepad(player, button, state)
    if self._games[player] then
        self._games[player]._gamepad:setButton(button, state)
    end
end

function gameHandler._defaultValues:keypressed(key, scancode, isrepeat)
    local player, virtualButton, state = self._inputHandler:keypressed(key, scancode, isrepeat)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function gameHandler._defaultValues:keyreleased(key, scancode)
    local player, virtualButton, state = self._inputHandler:keyreleased(key, scancode)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function gameHandler._defaultValues:gamepadpressed(joystick, button)
    local player, virtualButton, state = self._inputHandler:gamepadpressed(joystick, button)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function gameHandler._defaultValues:gamepadreleased(joystick, button)
    local player, virtualButton, state = self._inputHandler:gamepadreleased(joystick, button)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function gameHandler._defaultValues:prepareGames(players, settings)
    self._games = {}

    for i = 1, players do
        self._games[i] = excitingBrickGameGame.createGame(settings or self._gameSettings)
    end
end

function gameHandler._defaultValues:anyGamesPaused()
    return self._paused
end

function gameHandler._defaultValues:unpauseAllGames()
    for _, game in ipairs(self._games) do
        game:unpause()
    end
end

function gameHandler._defaultValues:pauseAllGames()
    for _, game in ipairs(self._games) do
        game:pause()
    end
end

function gameHandler._defaultValues:canUnpause()
    for _, game in ipairs(self._games) do
        if not game:canUnpause() then
            return false
        end
    end

    return true
end

-- TODO - Game mode should be able to give gamemode settings
function gameHandler.createGameHandler(data)
    data = data or {}

    local self = {}

    local theme = data.theme or "themes/default"
    local gameMode = data.gameMode or "modes/singleplayer/marathon"

    self._theme = type(theme) == "string" and excitingBrickGameTheme.loadTheme(theme) or theme
    self._gameMode = type(gameMode) == "string" and excitingBrickGameMode.loadMode(data.gameMode) or gameMode

    self._gameSettings = data._gameSettings or {}
    self._gameModeSettings = data._gameModeSettings or {}
    self._games = data.games
    self._paused = false

    self._inputHandler = data.inputHandler or inputHandler.createInputHandler()

    if not self._games then
        gameHandler._defaultValues.prepareGames(self, data.players or 1, self._gameSettings)
    end

    self._gameMode:init(self._games, self._gameModeSettings)

    return setmetatable(self, gameHandler._MT)
end

return gameHandler