local excitingBrickGameGarbage = {}

excitingBrickGameGarbage._MT = {}

function excitingBrickGameGarbage._MT:__index(key)
    if excitingBrickGameGarbage._defaultValues[key] then
        return excitingBrickGameGarbage._defaultValues[key]
    end

    return rawget(self, key)
end

excitingBrickGameGarbage._defaultValues = {}

-- Use lookup if it isn't empty, otherwise calculate
function excitingBrickGameGarbage._defaultValues:defaultComboFomula(combo)
    if self._comboLookup and #self._comboLookup > 0 then
        return self._comboLookup[combo] or self._comboMaxBonus
    end

    return math.min(math.floor(combo / 2), self._comboMaxBonus)
end

function excitingBrickGameGarbage._defaultValues:calculateGarbage(linesCleared, tSpin, awardBackToBack, combo, allClear)
    if allClear and self._allClearValue > 0 then
        return self._allClearValue
    end

    local baseValue = tSpin == "proper" and self._lineClearTSpinValues[linesCleared] or self._lineClearValues[linesCleared]

    local allClearBonus = allClear and self._allClearBonus or 0
    local comboBonus = self:_comboFormula(combo)
    local backToBackBonus = awardBackToBack and self._backToBackBonus or 0

    return baseValue + allClearBonus + comboBonus + backToBackBonus
end

excitingBrickGameGarbage._defaultValues.lineClearValues = {
    [0] = 0,
    0,
    1,
    2,
    4
}

excitingBrickGameGarbage._defaultValues.lineClearTSpinValues = {
    [0] = 0,
    2,
    4,
    6
}

excitingBrickGameGarbage._defaultValues.comboLookup = {
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    4
}

-- Defaults from Puyo Puyo Exciting!
excitingBrickGameGarbage._defaultValues.allClearValue = 10
excitingBrickGameGarbage._defaultValues.allClearBonus = 0
excitingBrickGameGarbage._defaultValues.backToBackBonus = 1
excitingBrickGameGarbage._defaultValues.comboMaxBonus = 5

function excitingBrickGameGarbage.createGarbageCalculator(data)
    data = data or {}

    local self = {}

    self._comboLookup = data.comboLookup or {}
    self._comboFormula = data.comboForuma or excitingBrickGameGarbage._defaultValues.defaultComboFomula
    self._comboMaxBonus = data.comboMaxBonus or excitingBrickGameGarbage._defaultValues.comboMaxBonus
    self._lineClearValues = data.lineClearValues or excitingBrickGameGarbage._defaultValues.lineClearValues
    self._lineClearTSpinValues = data.lineClearValues or excitingBrickGameGarbage._defaultValues.lineClearTSpinValues
    self._allClearValue = data._allClearValue or excitingBrickGameGarbage._defaultValues.allClearValue
    self._allClearBonus = data.allClearBonus or excitingBrickGameGarbage._defaultValues.allClearBonus
    self._backToBackBonus = data.backToBackBonus or excitingBrickGameGarbage._defaultValues.backToBackBonus

    return setmetatable(self, excitingBrickGameGarbage._MT)
end

return excitingBrickGameGarbage