local excitingBrickGameTheme = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_theme")
local inputHandler = require("mods").requireFromPlugin("input_handler")
local excitingBrickGameGameHandler = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_game_handler")
local excitingBrickGameSimpleMenu= require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_simple_menu")

local excitingBrickGameSceneHandler = {}

excitingBrickGameSceneHandler._defaultValues = {}

excitingBrickGameSceneHandler._MT = {}

function excitingBrickGameSceneHandler._MT:__index(key)
    if excitingBrickGameSceneHandler._defaultValues[key] then
        return excitingBrickGameSceneHandler._defaultValues[key]
    end

    return rawget(self, key)
end

excitingBrickGameSceneHandler._defaultValues = {}

function excitingBrickGameSceneHandler._defaultValues:draw()
    local theme = self._theme
    local gameHandler = self._gameHandler

    if gameHandler then
        local paused = gameHandler:anyGamesPaused()

        gameHandler:draw()

        if paused then
            theme:drawPauseMenu(self._pauseMenu:getDrawOptions(), self._pauseMenu._index)
        end

    else
        theme:drawMainMenu(self._mainMenu:getDrawOptions(), self._mainMenu._index)
    end
end

function excitingBrickGameSceneHandler._defaultValues:update(dt)
    if self._gameHandler then
        self._gameHandler:update(dt)
    end
end

function excitingBrickGameSceneHandler._defaultValues:startGameMode(gameMode)
    self._optionIndex = 1
    self._initData.gameMode = gameMode
    self._gameHandler = excitingBrickGameGameHandler.createGameHandler({
        players = 1,
        gameMode = gameMode,
        inputHandler = self._inputHandler,
        theme = self._theme,
    })
end

function excitingBrickGameSceneHandler._defaultValues:updateVirtualGamepad(player, button, state)
    if player and button  then
        local gameHandler = self._gameHandler

        if gameHandler then
            if gameHandler:anyGamesPaused() then
                self._pauseMenu:handleInputs(player, button, state)

            else
                gameHandler:updateVirtualGamepad(player, button, state)
            end

        else
            self._mainMenu:handleInputs(player, button, state)
        end
    end
end

function excitingBrickGameSceneHandler._defaultValues:keypressed(key, scancode, isrepeat)
    local player, virtualButton, state = self._inputHandler:keypressed(key, scancode, isrepeat)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function excitingBrickGameSceneHandler._defaultValues:keyreleased(key, scancode)
    local player, virtualButton, state = self._inputHandler:keyreleased(key, scancode)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function excitingBrickGameSceneHandler._defaultValues:gamepadpressed(joystick, button)
    local player, virtualButton, state = self._inputHandler:gamepadpressed(joystick, button)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function excitingBrickGameSceneHandler._defaultValues:gamepadreleased(joystick, button)
    local player, virtualButton, state = self._inputHandler:gamepadreleased(joystick, button)

    self:updateVirtualGamepad(player, virtualButton, state)
end

function excitingBrickGameSceneHandler.createScene(data)
    data = data or {}

    local self = {}

    self._initData = data

    local theme = data.theme or "themes/default"

    self._theme = type(theme) == "string" and excitingBrickGameTheme.loadTheme(theme) or theme
    self._inputHandler = data.inputHandler or inputHandler.createInputHandler()
    self._gameHandler = nil

    local mainMenuOptions = data.options or {
        {
            text = "Marathon",
            callback = function()
                self:startGameMode("modes/singleplayer/marathon")
            end,
        },
        {
            text = "Sprint",
            callback = function()
                self:startGameMode("modes/singleplayer/sprint")
            end,
        },
        {
            text = "Ultra",
            callback = function()
                self:startGameMode("modes/singleplayer/ultra")
            end,
        },
        {
            text = "Exit Game",
            callback = function()
                love.event.quit()
            end,
        },
    }

    local pauseMenuOptions = data.pauseOptions or {
        {
            text = "Resume",
            callback = function()
                self._gameHandler:unpauseAllGames()
            end,
            visible = function()
                return self._gameHandler:canUnpause()
            end
        },
        {
            text = "Retry",
            callback = function()
                -- TODO - This can fail if gameMode is not a stirng 
                local gameMode = self._initData.gameMode or "modes/singleplayer/marathon"

                self:startGameMode(gameMode)
            end,
        },
        {
            text = "Main Menu",
            callback = function()
                -- Clearing game handler is enough to get back to main menu
                self._gameHandler = nil
                self._initData.gameMode = nil
            end
        },
    }

    self._mainMenu = excitingBrickGameSimpleMenu.createMenu(mainMenuOptions)
    self._pauseMenu = excitingBrickGameSimpleMenu.createMenu(pauseMenuOptions)

    return setmetatable(self, excitingBrickGameSceneHandler._MT)
end

return excitingBrickGameSceneHandler
