local excitingBrickGameScore = {}

excitingBrickGameScore._MT = {}

function excitingBrickGameScore._MT:__index(key)
    if excitingBrickGameScore._defaultValues[key] then
        return excitingBrickGameScore._defaultValues[key]
    end

    return rawget(self, key)
end

excitingBrickGameScore._defaultValues = {}

function excitingBrickGameScore._defaultValues:getDropBonus(softDrop, hardDrop)
    if hardDrop then
        return self._hardDropValue

    elseif softDrop then
        return self._softDropValue

    else
        return 0
    end
end

function excitingBrickGameScore._defaultValues:calculateScore(level, linesCleared, tSpin, awardBackToBack, combo, allClear)
    local baseValue = tSpin == "proper" and self._lineClearTSpinValues[linesCleared] or self._lineClearValues[linesCleared]
    local comboBonus = math.min(self._comboMultiplier * combo, self._comboMaxBonus)
    local backToBackBonus = awardBackToBack and self._backToBackMultiplier or 1
    local miniTSpinBonus = tSpin == "mini" and self._miniTSpinBonus or 0
    local allClearBonus = allClear and self._allClearBonus[linesCleared] or 0

    return ((baseValue + miniTSpinBonus) * backToBackBonus + comboBonus + allClearBonus) * level
end

excitingBrickGameScore._defaultValues.lineClearValues = {
    [0] = 0,
    100,
    300,
    500,
    800
}

excitingBrickGameScore._defaultValues.allClearBonus = {
    [0] = 0,
    800,
    1000,
    1800,
    2000
}

excitingBrickGameScore._defaultValues.lineClearTSpinValues = {
    [0] = 400,
    800,
    1200,
    1600
}

excitingBrickGameScore._defaultValues.backToBackMultiplier = 1.5
excitingBrickGameScore._defaultValues.softDropValue = 1
excitingBrickGameScore._defaultValues.hardDropValue = 2
excitingBrickGameScore._defaultValues.comboMultiplier = 50
excitingBrickGameScore._defaultValues.comboMaxBonus = 1000
excitingBrickGameScore._defaultValues.miniTSpinBonus = 100

function excitingBrickGameScore.createScoreCalculator(data)
    data = data or {}

    local self = {}

    self._lineClearValues = data.lineClearValues or excitingBrickGameScore._defaultValues.lineClearValues
    self._lineClearTSpinValues = data.lineClearValues or excitingBrickGameScore._defaultValues.lineClearTSpinValues
    self._allClearBonus = data.allClearBonus or excitingBrickGameScore._defaultValues.allClearBonus
    self._backToBackMultiplier = data.lineClearValues or excitingBrickGameScore._defaultValues.backToBackMultiplier
    self._softDropValue = data.softDropValue or excitingBrickGameScore._defaultValues.softDropValue
    self._hardDropValue = data.hardDropValue or excitingBrickGameScore._defaultValues.hardDropValue
    self._comboMultiplier = data.comboMultiplier or excitingBrickGameScore._defaultValues.comboMultiplier
    self._comboMaxBonus = data.comboMaxBonus or excitingBrickGameScore._defaultValues.comboMaxBonus
    self._miniTSpinBonus = data.miniTSpinBonus or excitingBrickGameScore._defaultValues.miniTSpinBonus

    return setmetatable(self, excitingBrickGameScore._MT)
end

return excitingBrickGameScore