-- TODO - Add support for menu navigation bindings
-- TODO - Add support for detecting joysticks/keyboard

local inputHandler = {}

inputHandler._MT = {}

function inputHandler._MT:__index(key)
    if inputHandler._defaultValues[key] then
        return inputHandler._defaultValues[key]
    end

    return rawget(self, key)
end

inputHandler._defaultValues = {}

-- Keyboard key = virtual button
inputHandler._defaultValues._gameplayBindingsKeyboard = {
    left = "left",
    right = "right",
    up = "hardDrop",
    down = "softDrop",
    escape = "pause",
    z = "rotateLeft",
    w = "rotateLeft", -- French/Italian layout support
    y = "rotateLeft", -- German layout support
    x = "rotateRight",
    c = "hold",
}

-- Joystick button = virtual button
inputHandler._defaultValues._gameplayBindingsJoystick = {
    dpleft = "left",
    dpright = "right",
    dpup = "hardDrop",
    dpdown = "softDrop",
    start = "pause",
    a = "rotateLeft",
    b = "rotateRight",
    y = "rotateRight",
    x = "rotateLeft",
    leftshoulder = "hold",
    rightshoulder = "hold"
}

-- TODO - Player number on all input callbacks
function inputHandler._defaultValues:keypressed(key, scancode, isrepeat)
    local mapping = self._gameplayBindingsKeyboard[key]

    if mapping then
        return 1, mapping, true
    end
end

function inputHandler._defaultValues:keyreleased(key, scancode)
    local mapping = self._gameplayBindingsKeyboard[key]

    if mapping then
        return 1, mapping, false
    end
end

function inputHandler._defaultValues:gamepadpressed(joystick, button)
    local mapping = self._gameplayBindingsJoystick[button]

    if mapping then
        return 1, mapping, true
    end
end

function inputHandler._defaultValues:gamepadreleased(joystick, button)
    local mapping = self._gameplayBindingsJoystick[button]

    if mapping then
        return 1, mapping, false
    end
end

function inputHandler.createInputHandler(data)
    data = data or {}

    local self = {}

    self._gameplayBindingsKeyboard = data.bindingsKeyboard or inputHandler._defaultValues._gameplayBindingsKeyboard
    self._gameplayBindingsJoystick = data.bindingsJoystick or inputHandler._defaultValues._gameplayBindingsJoystick

    return setmetatable(self, inputHandler._MT)
end

return inputHandler
