-- TODO - Support TCP

local binarySerializer = require("mods").requireFromPlugin("binary_serialize")
local binaryWriter = require("mods").requireFromPlugin("binary_writer")
local binaryReader = require("mods").requireFromPlugin("binary_reader")
local socket = require("mods").requireFromPlugin("socket")

local networkClient = {}

networkClient._MT = {}

function networkClient._MT:__index(key)
    if networkClient._defaultValues[key] then
        return networkClient._defaultValues[key]
    end

    return rawget(self, key)
end

networkClient._defaultValues = {}

function networkClient._defaultValues:serialize(data)
    self._binaryWriter:clear()

    return binarySerializer.serialize(data, {}, self._binaryWriter)
end

function networkClient._defaultValues:send(message)
    local payload = self:serialize(message)

    return self._socket:send(payload)
end

function networkClient._defaultValues:unserialize(data)
    self._binaryReader:setContent(data)

    return binarySerializer.unserialize(nil, {}, self._binaryReader)
end

function networkClient._defaultValues:receive()
    local data, msg = self._socket:receive()

    if data then
        local unserialized = self:unserialize(data)

        return unserialized, msg

    else
        return data, msg
    end
end

function networkClient._defaultValues:getSocket()
    return self._socket
end

function networkClient.createUDPClient(address, port, timeout)
    local self = {}

    self._socket = socket.udp()
    self._socketType = "udp"

    self._socket:setpeername(address, port)
    self._socket:settimeout(timeout)

    self._binaryWriter = binaryWriter()
    self._binaryReader = binaryReader()

    return setmetatable(self, networkClient._MT)
end

return networkClient