-- TODO - Support TCP

local binarySerializer = require("mods").requireFromPlugin("binary_serialize")
local binaryWriter = require("mods").requireFromPlugin("binary_writer")
local binaryReader = require("mods").requireFromPlugin("binary_reader")
local socket = require("mods").requireFromPlugin("socket")

local networkServer = {}

networkServer._MT = {}

function networkServer._MT:__index(key)
    if networkServer._defaultValues[key] then
        return networkServer._defaultValues[key]
    end

    return rawget(self, key)
end

networkServer._defaultValues = {}

function networkServer._defaultValues:serialize(data)
    self._binaryWriter:clear()

    return binarySerializer.serialize(data, {}, self._binaryWriter)
end

function networkServer._defaultValues:send(message)
    local payload = self:serialize(message)

    return self._socket:send(payload)
end

function networkServer._defaultValues:sendto(message, ip, port)
    local payload = self:serialize(message)

    return self._socket:sendto(payload, ip, port)
end

function networkServer._defaultValues:unserialize(data)
    self._binaryReader:setContent(data)

    return binarySerializer.unserialize(nil, {}, self._binaryReader)
end

function networkServer._defaultValues:receive()
    local data, msg = self._socket:receive()

    if data then
        local unserialized = self:unserialize(data)

        return unserialized, msg

    else
        return data, msg
    end
end

function networkServer._defaultValues:receivefrom()
    local data, msg, port = self._socket:receivefrom()

    if data then
        self._binaryReader:setContent(data)
        local unserialized = self:unserialize(data)

        return unserialized, msg, port

    else
        return data, msg, port
    end
end

function networkServer._defaultValues:getSocket()
    return self._socket
end

function networkServer._defaultValues:getCurrentPacketNumber()
    return self._packetCounter
end

function networkServer._defaultValues:getNextPacketNumber()
    self._packetCounter = self._packetCounter + 1

    return self._packetCounter
end


function networkServer.createUDPServer(address, port, timeout)
    local self = {}

    self._socket = socket.udp()
    self._socketType = "udp"

    self._socket:setsockname(address, port)
    self._socket:settimeout(timeout)

    self._binaryWriter = binaryWriter()
    self._binaryReader = binaryReader()

    self._packetCounter = 0

    return setmetatable(self, networkServer._MT)
end

return networkServer