-- TODO - Handle simple log in

local packetHandlerManager = require("mods").requireFromPlugin("server.package_handler_manager")

local authentication = {}

authentication._MT = {}

function authentication._MT:__index(key)
    if authentication._defaultValues[key] then
        return authentication._defaultValues[key]
    end

    return rawget(self, key)
end

authentication._defaultValues = {}

authentication._defaultValues.timingFunction = os.clock

function authentication._defaultValues:update(dt, server)
    self:handleTimeout(dt, server)
end

function authentication._defaultValues:getIdentifier(ip, port)
    return string.format("%s:%s", ip, port)
end

-- Returns whether the user was added or not, followed by user data for the user
function authentication._defaultValues:addUser(ip, port)
    local identifier = self:getIdentifier(ip, port)

    if not self.users[identifier] then
        self.users[identifier] = {
            lastSeen = self:timingFunction(),
            ip = ip,
            port = port,
            name = "Guest"
        }

        self.userCount = self.userCount + 1

        print("Added user " .. identifier)

        return true, self.users[identifier]

    else
        self.users[identifier].lastSeen = self:timingFunction()

        return false, self.users[identifier]
    end
end

-- Returns whether the user was removed or not
function authentication._defaultValues:removeUser(identifier, port)
    identifier = port and self:getIdentifier(identifier, port) or identifier

    if self.users[identifier] then
        self.users[identifier] = nil

        self.userCount = self.userCount -1

        print("Removed user " .. identifier)

        return true, self.users[identifier]
    end

    return false
end

function authentication._defaultValues:handleTimeout(dt, server)
    self.timeoutAcc = self.timeoutAcc + dt

    if self.timeoutAcc > self.timeoutRate then
        self:timeoutUsers(server)

        self.timeoutAcc = self.timeoutAcc - self.timeoutRate
    end
end

function authentication._defaultValues:timeoutUsers(server)
    local now = self:timingFunction()

    for identifier, userData in pairs(self.users) do
        local lastKeepAlive = userData.lastKeepAlive or 0
        local lastSeen = userData.lastSeen

        if now > lastSeen + self.keepAliveDuration then
            if now > lastKeepAlive + self.keepAliveRate then
                self:sendKeepAlive(server, userData)
            end
        end

        if now > lastSeen + self.timeoutDuration then
            self:removeUser(identifier)
        end
    end
end

function authentication._defaultValues:getUserCount()
    return self.userCount
end

function authentication._defaultValues:sendKeepAlive(server, userData)
    local ip, port = userData.ip, userData.port

    userData.lastKeepAlive = self:timingFunction()

    server:sendto({"ping", server:getNextPacketNumber(), math.random(0, 1024)}, ip, port)
end

function authentication._defaultValues:sendWelcomeMessage(server, userData)
    local ip, port = userData.ip, userData.port

    server:sendto({"welcomeMessage", server:getNextPacketNumber(), "Welcome to test server :) Please play fair."}, ip, port)
end

function authentication._defaultValues:replyNameChangeRequest(server, ip, port, packetNumber, success)
    local payload = {
        success = success,
        repliesTo = packetNumber
    }

    if success then
        payload.message = "Name has been changed."

    else
        payload.message = "Name was not changed: "
    end

    server:sendto({"nameChangeReply", server:getNextPacketNumber(), payload}, ip, port)
end

authentication._defaultValues.packetHandler = {}

authentication._defaultValues.packetHandler[packetHandlerManager.handlerAnyPacketKey] = function(self, server, packetType, packetNumber, packetData, ip, port)
    local added, userData = self:addUser(ip, port)

    if added then
        self:sendWelcomeMessage(server, userData)
    end
end

function authentication._defaultValues.packetHandler.pong(server, packetNumber, packetData, ip, port)
    -- Stub method for now
    -- Registring that they are still "alive" is handled in the wildcard handler function
end

function authentication._defaultValues.packetHandler:changeName(server, packetNumber, packetData, ip, port)
    local success = false

    if type(packetData) == "string" and #packetData <= self.usernameByteLimit then
        local identifier = self:getIdentifier(ip, port)

        if self.users[identifier] then
            self.users[identifier].name = packetData
        end

        success = true
    end

    self:replyNameChangeRequest(server, ip, port, packetNumber, success)
end

function authentication.createHandler()
    local self = {}

    self.users = {}
    self.userCount = 0

    self.timeoutAcc = 0
    self.timeoutRate = 1
    self.timeoutDuration = 30

    self.keepAliveRate = 4
    self.keepAliveDuration = 10

    self.usernameByteLimit = 64

    return setmetatable(self, authentication._MT)
end

return authentication