local packetHandlerManager = require("mods").requireFromPlugin("server.package_handler_manager")
local networkServer = require("mods").requireFromPlugin("networking.network_server")

local address = LAUNCH_ARGUMENTS[2] or "0.0.0.0"
local bindPort = tonumber(LAUNCH_ARGUMENTS[3]) or 7777

local server = networkServer.createUDPServer(address, bindPort, 0)

local handlerNames = {
    "server.handlers.authentication",
    "server.handlers.simple_board_echo"
}

love.window.setTitle("Exciting Brick Game! - Server")

-- Test these and tweak
DRAW_INTERVAL = 5
UPDATE_INTERVAL = 1 / 144
MAIN_LOOP_SLEEP = 0.001

local packetsReceived = 0

-- Load and set up handlers by name
for _, name in ipairs(handlerNames) do
    packetHandlerManager.loadHandler(name)
end

local authenticationHandler = packetHandlerManager.getHandler("server.handlers.authentication")

local function handleEvents(server)
    -- Handle all waiting messages
    while true do
        local data, ip, port = server:receivefrom()

        if data then
            local packetType, packetNumber, packetData = data[1], data[2], data[3]

            packetHandlerManager.sendEvent(server, packetType, packetNumber, packetData, ip, port)

            packetsReceived = packetsReceived + 1

        else
            break
        end
    end
end

print("Starting server at ", address, bindPort)

function love.update(dt)
    handleEvents(server)

    packetHandlerManager.update(dt, server)
end

function love.draw()
    love.graphics.print(string.format("Packets %s", packetsReceived), 20, 20)
    love.graphics.print(string.format("FPS %s", love.timer.getFPS()), 20, 40)
    love.graphics.print(string.format("Users %s", authenticationHandler:getUserCount()), 20, 60)
end

function love.keyreleased(key, scancode)

end