local manager = {}

manager.handlers = {}

manager.handlerAnyPacketKey = "__anyPacket"

function manager.loadHandler(name, verbose)
    if verbose ~= false then
        print(string.format("! Loaded handler '%s'", name))
    end

    manager.handlers[name] = require("mods").requireFromPlugin(name).createHandler()
end

function manager.getHandler(name)
    return manager.handlers[name]
end

function manager.sendEvent(server, packetType, packetNumber, packetData, ip, port)
    local handlers = manager.handlers
    local anyPacketKey = manager.handlerAnyPacketKey

    for handlerName, handler in pairs(manager.handlers) do
        local packetHandler = handler.packetHandler

        local directHandler = packetHandler and packetHandler[packetType]
        local wildcardHandler = packetHandler and packetHandler[anyPacketKey]

        if directHandler then
            directHandler(handler, server, packetNumber, packetData, ip, port)
        end

        if wildcardHandler then
            wildcardHandler(handler, server, packetType, packetNumber, packetData, ip, port)
        end
    end
end

function manager.update(dt, server)
    for handlerName, handler in pairs(manager.handlers) do
        if handler.update then
            handler:update(dt, server)
        end
    end
end

return manager