-- TODO - Better ghost piece

local utils = require("mods").requireFromPlugin("utils")
local colors = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_default_colors")
local excitingBrickShapes = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_default_exciting_brick_shapes")
local excitingBrickShapeStruct = require("mods").requireFromPlugin("structs.exciting_brick_shape")

local theme = {}

local tileWidth = 16
local tileHeight = 16

local tileQuadLookup = {
    g = 1,
    i = 2,
    o = 3,
    t = 4,
    s = 5,
    z = 6,
    j = 7,
    l = 8,
    [" "] = 9
}

local previewexcitingBrickShapes = {}

for name, data in pairs(excitingBrickShapes) do
    previewexcitingBrickShapes[name] = excitingBrickShapeStruct.createexcitingBrickShape(data)
end

local useColoredRectangles = false
local useMinoTinting = false

local previewSpacing = tileHeight * 7

local ghostexcitingBrickShapeAlpha = 0.25

local defaultUnknownColor = {1.0, 1.0, 1.0}

local holdPreviewX = 0
local holdPreviewY = 0.75 * tileHeight

local playfieldX = tileWidth * 3
local playfieldY = 0
local playfieldScale = 2

local previewsX = (playfieldX + 11.5 * tileWidth) * playfieldScale
local previewsY = 0.75 * tileHeight + 1

local gameModeInfoX = previewsX + previewSpacing
local gameModeInfoY = tileHeight * 1.5
local gameModeInfoColor = {0 / 255, 45 / 255, 114 / 255}

local infoBoxTileWidth = 14
local infoBoxTileHeight = 8

local mainMenuEntryOffset = tileHeight * 2
local mainMenuButtonColor = {18 / 255, 0 / 255, 88 / 255}
local mainMenuButtonSelectedColor = {32 / 255, 0 / 255, 152 / 255}
local mainMenuXOffset = 15 * tileWidth
local mainMenuControlsX = 10 * tileWidth
local mainMenuControlsYOffset = 32.5 * tileHeight
local mainMenuControlsBoxTileHeight = 6

local unfocusedBoxColor = {0 / 255, 62 / 255, 125 / 255}
local unfocusedBoxTileHeight = 14

local clearInfoColor = {0 / 255, 83 / 255, 125 / 255}

local defaultTextColor = {1.0, 1.0, 1.0}

local comboX = 0
local comboY = holdPreviewY + 8 * tileHeight
local comboDuration = 2

local clearNameX = 0
local clearNameY = comboY + 4.5 * tileHeight
local clearNameDuration = 2

local backToBackX = 0
local backToBackY = clearNameY + 4.5 * tileHeight
local backToBackDuration = 2

local controlsYOffset = 38 * tileHeight

local scoreX = tileWidth * 8 * playfieldScale
local scoreY = tileHeight * 21.5 * playfieldScale

function theme:init(directory)
    local _, pluginInfo = require("mods").findLoadedMod("ProcrastinationHelper")
    directory = string.format("%s/%s", pluginInfo._mountPointLoenn, directory)

    self._tilesImage = love.graphics.newImage(directory .. "/tiles.png")
    self._tilesQuads = utils.tileImageToQuadMatrix(self._tilesImage, tileWidth, tileHeight)

    self._boardBorderImage = love.graphics.newImage(directory .. "/board_border.png")
    self._boardBorderQuads = utils.tileImageToQuadMatrix(self._boardBorderImage, tileWidth, tileHeight)

    self._buttonImage = love.graphics.newImage(directory .. "/button.png")
    self._buttonQuads = utils.tileImageToQuadMatrix(self._buttonImage, tileWidth, tileHeight)

    self._logoImage = love.graphics.newImage(directory .. "/logo.png")

    self._fontScale = 1
end

function theme:drawMinoColoredRectangle(x, y, mino, alpha)
    if colors[mino] then
        local pr, pg, pb, pa = love.graphics.getColor()
        local r, g, b = unpack(colors[mino])

        love.graphics.setColor(r, g, b, alpha)
        love.graphics.rectangle("fill", x, y, tileWidth, tileHeight)
        love.graphics.setColor(pr, pg, pb, pa)
    end
end

function theme:drawMinoImage(x, y, mino, alpha)
    local lookup = useMinoTinting and 1 or (tileQuadLookup[mino] or 9)
    local color = useMinoTinting and colors[mino] or defaultUnknownColor

    local quad = self._tilesQuads:get(lookup, 1)

    local pr, pg, pb, pa = love.graphics.getColor()
    local r, g, b = unpack(color)

    love.graphics.setColor(r, g, b, alpha)
    love.graphics.draw(self._tilesImage, quad, x, y)
    love.graphics.setColor(pr, pg, pb, pa)
end

function theme:drawMino(x, y, mino, alpha)
    if useColoredRectangles then
        self:drawMinoColoredRectangle(x, y, mino, alpha)

    else
        self:drawMinoImage(x, y, mino, alpha)
    end
end

-- TODO - Clip minos outside of grid
function theme:drawMinos(game)
    local playfield = game._playfield

    love.graphics.push()

    love.graphics.translate(tileWidth, tileHeight)

    local width, height = playfield._width, playfield._height
    local offsetX, offsetY = (playfield._width - playfield._visibleWidth) * tileWidth, (playfield._height - playfield._visibleHeight) * tileHeight

    local drawX, drawY = 0, 0
    local mino

    for x = 1, width do
        drawX = (x - 1) * tileWidth - offsetX

        for y = 1, height do
            drawY = (y - 1) * tileHeight - offsetY
            mino = playfield._playfieldMatrix:get(x, y)

            self:drawMino(playfieldX + drawX, playfieldY + drawY, mino)
        end
    end

    love.graphics.pop()
end

function theme:drawBorder(game)
    local playfield = game._playfield
    local width, height = playfield._visibleWidth, playfield._visibleHeight

    utils.drawNinePatch(self._boardBorderImage, self._boardBorderQuads, playfieldX, playfieldY, width + 2, height + 2, tileWidth, tileHeight, true)
end

function theme:drawexcitingBrickShapePreview(excitingBrickShape)
    local matrix

    if previewexcitingBrickShapes[excitingBrickShape] then
        matrix = previewexcitingBrickShapes[excitingBrickShape].rotations[1]

    elseif excitingBrickShape then
        matrix = excitingBrickShape.rotations[1]
    end

    if matrix then
        local width, height = matrix:size()

        for x = 1, width do
            for y = 1, height do
                self:drawMino(x * tileWidth, y * tileHeight, matrix:get(x, y, " "))
            end
        end
    end
end

function theme:drawexcitingBrickShape(game, excitingBrickShape, alpha)
    if excitingBrickShape then
        local playfield = game._playfield
        local matrix = excitingBrickShape:getMatrix()
        local width, height = matrix:size()

        local offsetX = playfieldX + (excitingBrickShape.x - excitingBrickShape.centerX - (playfield._width - playfield._visibleWidth) + 1) * tileWidth
        local offsetY = playfieldY + (excitingBrickShape.y - excitingBrickShape.centerY - (playfield._height - playfield._visibleHeight) + 1) * tileHeight

        for x = 1, width do
            for y = 1, height do
                self:drawMino((x - 1) * tileWidth + offsetX, (y - 1) * tileHeight + offsetY, matrix:get(x, y, " "), alpha)
            end
        end
    end
end

function theme:drawPreview(excitingBrickShape, onlyBackground)
    utils.drawNinePatch(self._boardBorderImage, self._boardBorderQuads, 0, 0, 7, 7, tileWidth, tileHeight, true)

    if onlyBackground ~= true then
        self:drawexcitingBrickShapePreview(excitingBrickShape)
    end
end

function theme:drawPreviews(game, settings)
    local offsetX, offsetY = previewsX, previewsY
    local previews = math.min(settings.previews or 5, #game._pieceBag._piecePreviews)
    local paused = game._paused

    love.graphics.push()
    love.graphics.translate(offsetX, offsetY)

    for i = 1, previews do
        local excitingBrickShape = game._pieceBag._piecePreviews[i]

        self:drawPreview(excitingBrickShape, paused)

        love.graphics.translate(0, previewSpacing)
    end

    love.graphics.pop()
end

function theme:drawHold(game, settings)
    love.graphics.push()
    love.graphics.translate(holdPreviewX, holdPreviewY)

    local excitingBrickShape = game._heldexcitingBrickShape
    local paused = game._paused

    self:drawPreview(excitingBrickShape, paused)

    love.graphics.pop()
end

function theme:drawPlayfield(game)
    self:drawBorder(game)
    self:drawMinos(game)
end

function theme:getLastClear(game)
    local infos = game:getLineClearInformations()
    local lastClear = infos[#infos]

    return lastClear
end

function theme:drawBox(x, y, text, image, quads, color, fontSize, boxTileWidth, boxTileHeight)
    fontSize = fontSize or self._fontScale

    boxTileWidth = boxTileWidth or 7
    boxTileHeight = boxTileHeight or 4

    love.graphics.push()
    love.graphics.translate(x, y)

    local textType = type(text)

    if textType == "function" then
        text = text()
    end

    if textType == "string" then
        local width = boxTileWidth * tileWidth
        local height = boxTileHeight * tileHeight

        utils.drawNinePatch(image, quads, 0, 0, boxTileWidth, boxTileHeight, tileWidth, tileHeight, true, color)
        utils.printCenteredText(text, 0, 0, width, height, nil, fontSize, defaultTextColor)

    else
        love.graphics.draw(text)
    end

    love.graphics.pop()
end

function theme:drawTextBox(x, y, text)
    self:drawBox(x, y, text, self._boardBorderImage, self._boardBorderQuads)
end

function theme:drawButton(x, y, text, color, fontSize, boxTileWidth, boxTileHeight)
    self:drawBox(x, y, text, self._buttonImage, self._buttonQuads, color, fontSize, boxTileWidth, boxTileHeight)
end

function theme:drawTemporaryInfo(game, clear, x, y, duration, text, color)
    if clear then
        color = color or clearInfoColor

        if game:getGameDuration() - clear.time < duration then
            self:drawButton(x, y, text, color)
        end
    end
end

function theme:drawComboCounter(game, clear)
    if clear and clear.combo > 0 then
        self:drawTemporaryInfo(game, clear, comboX, comboY, comboDuration, string.format("Combo: %d", clear.combo))
    end
end

function theme:drawClearName(game, clear)
    if clear then
        self:drawTemporaryInfo(game, clear, clearNameX, clearNameY, clearNameDuration, clear.name)
    end
end

function theme:drawBackToBack(game, clear)
    if clear and clear.backToBack then
        self:drawTemporaryInfo(game, clear, backToBackX, backToBackY, backToBackDuration, "Back To Back!")
    end
end

function theme:drawScorebox(game)
    love.graphics.push()

    love.graphics.translate(scoreX, scoreY)
    love.graphics.scale(self._fontScale, self._fontScale)
    love.graphics.print(string.format("%08d", game._score), 0, 0)

    love.graphics.pop()
end

local controlsText = "Controls:\nArrow keys\n" .. string.upper(love.keyboard.getKeyFromScancode("z") .. " x c")

function theme:drawGameModeInfoBoxes(gameMode, game)
    local gameModeInfoBoxes = {}

    if gameMode.getInfoBoxes then
        gameModeInfoBoxes = gameMode:getInfoBoxes(game)
    end

    local fontSize = self._fontScale
    local x, y = gameModeInfoX, gameModeInfoY
    local widest = 0

    for _, text in ipairs(gameModeInfoBoxes) do
        local textWidth = utils.getTextSize(text, nil, fontSize)

        widest = math.max(widest, textWidth)
    end

    for _, text in ipairs(gameModeInfoBoxes) do
        local _, textHeight = utils.getTextSize(text, nil, fontSize)
        local boxTileWidth = math.ceil(widest / tileWidth) + 2
        local boxTileHeight = math.ceil(textHeight / tileHeight) + 2

        self:drawButton(x, y, text, gameModeInfoColor, fontSize, boxTileWidth, boxTileHeight)

        y = y + boxTileHeight * tileHeight + tileHeight
    end

    do
        local _, textHeight = utils.getTextSize(controlsText, nil, fontSize)
        local boxTileWidth = math.ceil(widest / tileWidth) + 2
        local boxTileHeight = math.ceil(textHeight / tileHeight) + 2

        self:drawButton(gameModeInfoX, gameModeInfoY + controlsYOffset, controlsText, gameModeInfoColor, fontSize, boxTileWidth, boxTileHeight)
    end
end

function theme:drawGame(gameMode, game, settings)
    settings = settings or {}

    love.graphics.push()
    love.graphics.scale(playfieldScale, playfieldScale)

    self:drawPlayfield(game, settings)
    self:drawexcitingBrickShape(game, game._ghostexcitingBrickShape, ghostexcitingBrickShapeAlpha)
    self:drawexcitingBrickShape(game, game._currentexcitingBrickShape)

    love.graphics.pop()

    self:drawPreviews(game, settings)
    self:drawHold(game, settings)

    -- Line clear information
    local lastClear = self:getLastClear(game)

    self:drawComboCounter(game, lastClear)
    self:drawClearName(game, lastClear)
    self:drawBackToBack(game, lastClear)

    self:drawGameModeInfoBoxes(gameMode, game)

    self:drawScorebox(game)
end

function theme:drawMenu(options, optionIndex, buttonColor, buttonSelectedColor, xOffset)
    buttonColor = buttonColor or mainMenuButtonColor
    buttonSelectedColor = buttonSelectedColor or mainMenuButtonSelectedColor
    xOffset = xOffset or 0

    local windowWidth, windowHeight = love.graphics.getDimensions()

    local totalHeight = (infoBoxTileHeight * tileHeight + mainMenuEntryOffset) * #options - mainMenuEntryOffset

    local x = math.floor((windowWidth - infoBoxTileWidth * tileWidth) / 2) + xOffset
    local y = math.floor((windowHeight - totalHeight) / 2)

    local fontSize = 2 * self._fontScale

    for i, option in ipairs(options) do
        local color = i == optionIndex and buttonSelectedColor or buttonColor

        self:drawButton(x, y, option.text, color, fontSize, infoBoxTileWidth, infoBoxTileHeight)

        y = y + infoBoxTileHeight * tileHeight + mainMenuEntryOffset
    end
end

function theme:drawMenuControls()
    local fontSize = 2 * self._fontScale
    local textWidth = utils.getTextSize(controlsText, nil, fontSize)
    textWidth = math.ceil(textWidth / tileWidth) + 2

    self:drawButton(mainMenuControlsX, gameModeInfoY + mainMenuControlsYOffset, controlsText, gameModeInfoColor, fontSize, textWidth, mainMenuControlsBoxTileHeight)
end

function theme:drawLogo()
    love.graphics.draw(self._logoImage, 2 * tileWidth, 2 * tileHeight)
end

function theme:drawMainMenu(options, optionIndex)
    self:drawLogo()
    self:drawMenuControls()
    self:drawMenu(options, optionIndex, nil, nil, mainMenuXOffset)
end

function theme:drawPauseMenu(options, optionIndex)
    self:drawMenu(options, optionIndex)
end

function theme:drawUnfocusedMessage()
    local text = "CLICK ON WINDOW\nTO FOCUS"
    local fontSize = 4 * self._fontScale

    local windowWidth, windowHeight = love.graphics.getDimensions()

    local textWidth = utils.getTextSize(text, nil, fontSize)
    textWidth = math.ceil(textWidth / tileWidth) + 2
    local textHeight = unfocusedBoxTileHeight

    local x = math.floor((windowWidth - textWidth * tileWidth) / 2)
    local y = math.floor((windowHeight - textHeight * tileHeight) / 2)

    self:drawButton(x, y, text, unfocusedBoxColor, fontSize, textWidth, textHeight)
end

return theme
