local virtualGamepad = {}

virtualGamepad._MT = {}

function virtualGamepad._MT:__index(key)
    if virtualGamepad._defaultValues[key] then
        return virtualGamepad._defaultValues[key]
    end

    return rawget(self, key)
end

virtualGamepad._defaultValues = {}

function virtualGamepad._defaultValues:setButton(button, state)
    local previousState = self._buttons[button]

    if previousState ~= state then
        self._buttons[button] = state
        self._buttonChangedAt[button] = self._timingFunction()

        if state then
            self:_buttonPressed(button)

        else
            self:_buttonReleased(button)
        end

        self:_buttonChanged(button, state)
    end
end

function virtualGamepad._defaultValues:getButton(button)
    return self._buttons[button]
end

function virtualGamepad._defaultValues:getButtonStateStart(button)
    return self._buttonsChangedAt[button] or -1
end

function virtualGamepad._defaultValues:getButtonStateDuration(button)
    if self._buttonsChangedAt[button] then
        return self._buttonsChangedAt[button] - self._timingFunction
    end

    return -1
end

virtualGamepad._defaultValues._defaultButtons = {
    "left", "right", "hardDrop", "softDrop",
    "hold", "rotateRight", "rotateLeft",
    "pause"
}

function virtualGamepad.createGamepad(data)
    data = data or {}

    local self = {}

    self._buttons = {}
    self._buttonChangedAt = {}

    self._buttonPressed = data.buttonPressed or function() end
    self._buttonReleased = data.buttonReleased or function() end
    self._buttonChanged = data.buttonChanged or function() end

    self._timingFunction = data.timingFunction or love.timer.getTime

    for _, button in ipairs(virtualGamepad._defaultValues._defaultButtons) do
        self._buttons[button] = false
        self._buttonChangedAt[button] = self._timingFunction()
    end

    return setmetatable(self, virtualGamepad._MT)
end

return virtualGamepad